from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Any

import pygls.lsp.types as lsp_types
import yaramod

from yls import utils

log = logging.getLogger(__name__)

# Keywords which can occur in condition section
CONDITION_KEYWORDS = {
    "all",
    "and",
    "any",
    "at",
    "contains",
    "endswith",
    "entrypoint",
    "false",
    "filesize",
    "for",
    "icontains",
    "iendswith",
    "iequals",
    "in",
    "int16",
    "int16be",
    "int32",
    "int32be",
    "int8",
    "int8be",
    "istartswith",
    "matches",
    "none",
    "not",
    "of",
    "or",
    "startswith",
    "them",
    "true",
    "uint16",
    "uint16be",
    "uint32",
    "uint32be",
    "uint8",
    "uint8be",
    "defined",
}


@dataclass
class Symbol:
    name: str

    def to_completion_items(self) -> list[lsp_types.CompletionItem]:
        return []

    # pylint: disable-next=unused-argument
    def to_hover(self, path: str | None = None) -> lsp_types.Hover:
        return lsp_types.Hover(contents=utils.markdown_content(self.name))

    def kind(self) -> lsp_types.CompletionItemKind:
        return lsp_types.CompletionItemKind.Text


@dataclass
class Structure(Symbol):
    name: str
    attributes: dict[str, Any]

    def detail(self) -> str:
        return "struct"

    def kind(self) -> lsp_types.CompletionItemKind:
        return lsp_types.CompletionItemKind.Struct

    def to_completion_items(self) -> list[lsp_types.CompletionItem]:
        log.debug(f'[to_completion_items] Creating completion for structure "{self.name}"')
        return [
            lsp_types.CompletionItem(
                label=self.name,
                kind=self.kind(),
                insert_text=self.name,
                insert_text_format=lsp_types.InsertTextFormat.PlainText,
                sort_text="a",  # High priority
            )
        ]


@dataclass
class Reference(Symbol):
    name: str
    symbol: Symbol

    def detail(self) -> str:
        return f"ref to {self.symbol.name}"

    def kind(self) -> lsp_types.CompletionItemKind:
        return self.symbol.kind()

    def to_completion_items(self) -> list[lsp_types.CompletionItem]:
        log.debug(f'[to_completion_items] Creating completion for reference "{self.name}"')
        return [
            lsp_types.CompletionItem(
                label=self.name,
                kind=self.kind(),
                insert_text=self.name,
                insert_text_format=lsp_types.InsertTextFormat.PlainText,
                detail=self.detail(),
                sort_text="a",  # High priority
            )
        ]


@dataclass
class Value(Symbol):
    name: str
    type_: yaramod.ExpressionType
    documentation: str = ""

    def detail(self) -> str:
        return Function.TYPE_TO_STRING[self.type_]

    def is_constant(self) -> bool:
        return self.name.isupper()

    def kind(self) -> lsp_types.CompletionItemKind:
        return (
            lsp_types.CompletionItemKind.Constant
            if self.is_constant()
            else lsp_types.CompletionItemKind.Variable
        )

    def sort_text(self) -> str:
        return "ab" if self.is_constant() else "a"

    def to_completion_items(self) -> list[lsp_types.CompletionItem]:
        log.debug(f'[to_completion_items] Creating completion for variable "{self.name}"')
        return [
            lsp_types.CompletionItem(
                label=self.name,
                kind=self.kind(),
                insert_text_format=lsp_types.InsertTextFormat.PlainText,
                documentation=utils.markdown_content(self.documentation),
                detail=self.detail(),
                sort_text=self.sort_text(),
            )
        ]

    def signature(self) -> str:
        return f"{self.name}:{self.detail()}"

    def hover_string(self) -> str:
        res = f"**{self.signature()}**\n********\n\n{self.documentation}"
        return res

    def to_hover(self, path: str | None = None) -> lsp_types.Hover:
        return lsp_types.Hover(contents=utils.markdown_content(self.documentation))


@dataclass
class Overload:
    parameters: list[yaramod.ExpressionType]
    documentation: str

    def param_string(self) -> str:
        return ", ".join(map(lambda p: Function.TYPE_TO_STRING[p], self.parameters))


@dataclass
class Function(Symbol):
    name: str
    overloads: list[Overload]
    return_type: yaramod.ExpressionType
    documentation: str = ""

    TYPE_TO_STRING = {
        yaramod.ExpressionType.Undefined: "undefined",
        yaramod.ExpressionType.Bool: "bool",
        yaramod.ExpressionType.Int: "int",
        yaramod.ExpressionType.String: "string",
        yaramod.ExpressionType.Regexp: "regex",
        yaramod.ExpressionType.Object: "object",
        yaramod.ExpressionType.Float: "float",
    }

    TYPE_TO_SHORT_STRING = {
        yaramod.ExpressionType.Undefined: "u",
        yaramod.ExpressionType.Bool: "b",
        yaramod.ExpressionType.Int: "i",
        yaramod.ExpressionType.String: "s",
        yaramod.ExpressionType.Regexp: "r",
        yaramod.ExpressionType.Object: "o",
        yaramod.ExpressionType.Float: "f",
    }

    def snippet(self) -> str:
        # return "{}(${{0:params}})".format(name)
        return f"{self.name}($0)"

    def signature(self) -> str:
        return f"{self.name}({self.overloads[0].param_string()}) -> {self.TYPE_TO_STRING[self.return_type]}"

    def hover_documentation(self) -> str:
        res = ""
        # NOTE: Handle overloads
        overload = self.overloads[0]

        res += f"**{self.signature()}**\n********\n\n"

        documentation = overload.documentation
        for line in documentation.splitlines(keepends=True):
            if line.strip().casefold().startswith("example"):
                res += "```\n"
                res += line
                res += "\n```\n"
            else:
                res += line

        return res

    def to_completion_items(self) -> list[lsp_types.CompletionItem]:
        log.debug(f'[to_completion_items] Creating completion for function "{self.name}"')
        res = []
        for overload in self.overloads:
            res.append(
                lsp_types.CompletionItem(
                    label=self.name,
                    kind=lsp_types.CompletionItemKind.Function,
                    insert_text=self.snippet(),
                    insert_text_format=lsp_types.InsertTextFormat.Snippet,
                    command=lsp_types.Command(
                        title="editor.action.triggerParameterHints",
                        command="editor.action.triggerParameterHints",
                        arguments=[],
                    ),
                    documentation=utils.markdown_content(overload.documentation),
                    detail=overload.param_string(),
                    sort_text="a",  # High priority
                )
            )
        return res

    def to_signature_information(
        self, path: str | None = None
    ) -> list[lsp_types.SignatureInformation]:
        res = []
        for overload in self.overloads:
            path = path or self.name
            doc = utils.markdown_content(overload.documentation)

            signature_string = (
                f"{path}({overload.param_string()}) -> {self.TYPE_TO_STRING[self.return_type]}"
            )
            parameters = []
            for param in overload.parameters:
                parameters.append(lsp_types.ParameterInformation(label=self.TYPE_TO_STRING[param]))

            res.append(
                lsp_types.SignatureInformation(
                    label=signature_string, parameters=parameters, documentation=doc
                )
            )

        return res

    def to_hover(self, path: str | None = None) -> lsp_types.Hover:
        return lsp_types.Hover(contents=utils.markdown_content(self.documentation))


@dataclass
class Array(Symbol):
    name: str
    element: Structure | Reference | Value
    documentation: str = ""

    def detail(self) -> str:
        return self.element.detail()

    def snippet(self) -> str:
        return f"{self.name}[$0]"

    def to_completion_items(self) -> list[lsp_types.CompletionItem]:
        log.debug(f'[to_completion_items] Creating completion for array "{self.name}"')
        return [
            lsp_types.CompletionItem(
                label=self.name,
                kind=self.element.kind(),
                insert_text=self.snippet(),
                insert_text_format=lsp_types.InsertTextFormat.Snippet,
                documentation=utils.markdown_content(self.documentation),
                detail=self.detail(),
                sort_text="a",  # High priority
            )
        ]


@dataclass
class Dictionary(Symbol):
    name: str
    element: Structure | Reference | Value
    documentation: str = ""

    def detail(self) -> str:
        return self.element.detail()

    def snippet(self) -> str:
        return f'{self.name}["$0"]'

    def to_completion_items(self) -> list[lsp_types.CompletionItem]:
        log.debug(f'[to_completion_items] Creating completion for dictionary "{self.name}"')
        return [
            lsp_types.CompletionItem(
                label=self.name,
                kind=self.element.kind(),
                insert_text=self.snippet(),
                insert_text_format=lsp_types.InsertTextFormat.Snippet,
                documentation=utils.markdown_content(self.documentation),
                detail=self.detail(),
                sort_text="a",  # High priority
            )
        ]


class CompletionCache:
    modules: Structure

    def __init__(self) -> None:
        pass

    def get_symbols_matching(self, path: str) -> list[Symbol]:
        """Returns a list of symbols that match the given path.

        Examples:
        - `pe.MACHINE_ARM` returns 3 symbols that match the prefix
          (MACHINE_ARM, MACHINE_ARM64, MACHINE_ARMNT).
        - `p` returns 2 symbols that match the prefix (pe, phish).
        - `` (empty string) returns all module symbols.
        """
        symbols, path_remainder = self._get_symbols(path)

        # Construct a list of results that match the given prefix
        symbols = list(filter(lambda x: x.name.startswith(path_remainder), symbols))

        return symbols

    def _get_symbols(self, path: str) -> tuple[list[Symbol], str]:
        """Return list of symbols that match the longest prefix path and a path remainder.

        If the match is exact (e.g. `pe.MACHINE_ARM64`) the returned remained is
        empty string. Otherwise the remainder is the first part that did not match
        the path.
        """

        # Split the path
        parts = path.split(".")

        # Initialize the search
        finder = self.modules

        # Enumerate all parts except the last one. This is because we want to find
        # all symbols with given prefix and manually select only those that match
        # the remainder.
        for i, part in enumerate(parts[:-1]):
            try:
                finder = finder.attributes[part]
            except KeyError:
                # The path is invalid
                return list(finder.attributes.values()), ".".join(parts[i:])
            except AttributeError:
                # We hit the "end" (Value, Array, ...)
                return [], ".".join(parts[i:])
        try:
            return list(finder.attributes.values()), parts[-1]
        except AttributeError:
            # In case we finish on a non Structure symbol
            # Example: 'pe.MACHINE_ARM64.'
            return [], ""

    def get_symbol(self, path: str) -> Symbol | None:
        """Returns a symbols that exactly matches the specified path.

        If the symbol is not found returns None.
        """
        parts = path.split(".")
        finder = self.modules
        for part in parts:
            try:
                finder = finder.attributes[part]
            except (KeyError, AttributeError):
                # We either tried to advance with invalid path or we
                # tried to move after we reached the "end". For example
                # selecting a Value and trying to advance past it.
                return None
        return finder

    def get_function(self, path: str) -> Function | None:
        symbol = self.get_symbol(path)
        if symbol is None:
            return None

        if not isinstance(symbol, Function):
            return None

        return symbol

    @staticmethod
    def visit_symbol(symbol: Any) -> Symbol:
        # pylint: disable=no-else-return,too-many-branches,too-many-return-statements
        if symbol.is_structure:
            attributes = {}
            for name, nested_symbol in symbol.attributes.items():
                attributes[name] = CompletionCache.visit_symbol(nested_symbol)
            return Structure(symbol.name, attributes)
        elif symbol.is_reference:
            return Reference(symbol.name, symbol.symbol)
        elif symbol.is_function:
            overloads = []
            for parameters, doc in zip(symbol.overloads, symbol.documentations):
                overloads.append(Overload(parameters, doc))
            return Function(symbol.name, overloads, symbol.return_type, symbol.documentation)
        elif symbol.is_array:
            element: Structure | Reference | Value
            if symbol.element_type != yaramod.ExpressionType.Object:
                # This array has a simple type
                # All are from dotnet module
                element = Value(symbol.name, symbol.element_type)
            else:
                attributes = {}
                if symbol.structure.is_reference:
                    element = Reference(symbol.name, symbol.structure.symbol)
                else:
                    for name, nested_symbol in symbol.structure.attributes.items():
                        attributes[name] = CompletionCache.visit_symbol(nested_symbol)
                    element = Structure(symbol.name, attributes)
            return Array(symbol.name, element, symbol.documentation)
        elif symbol.is_dictionary:
            element = Value(symbol.name, symbol.element_type)
            return Dictionary(symbol.name, element, symbol.documentation)
        elif symbol.is_value:
            return Value(symbol.name, symbol.data_type, symbol.documentation)
        else:
            log.warning(
                f"[visit_symbol] Unimplemented type {symbol} '{symbol.name}' creating a default fallback."
            )
            return Value(symbol.name, symbol.data_type)

    @staticmethod
    def from_yaramod(ymod: yaramod.Yaramod) -> "CompletionCache":
        res = CompletionCache()

        attributes = {}
        for name, module in ymod.modules.items():
            attributes[name] = CompletionCache.visit_symbol(module.structure)

        res.modules = Structure("", attributes)

        return res



from __future__ import annotations

import functools
import logging
import pathlib
import re
from typing import Any

import pygls.lsp.types as lsp_types
import yaramod
from pygls.lsp.methods import CODE_ACTION
from pygls.lsp.methods import CODE_LENS
from pygls.lsp.methods import COMPLETION
from pygls.lsp.methods import DEFINITION
from pygls.lsp.methods import DOCUMENT_HIGHLIGHT
from pygls.lsp.methods import DOCUMENT_SYMBOL
from pygls.lsp.methods import FORMATTING
from pygls.lsp.methods import HOVER
from pygls.lsp.methods import INITIALIZED
from pygls.lsp.methods import REFERENCES
from pygls.lsp.methods import SIGNATURE_HELP
from pygls.lsp.methods import TEXT_DOCUMENT_DID_CHANGE
from pygls.lsp.methods import TEXT_DOCUMENT_DID_OPEN
from pygls.lsp.methods import TEXT_DOCUMENT_DID_SAVE
from pygls.lsp.methods import WORKSPACE_DID_CHANGE_CONFIGURATION
from pygls.server import LanguageServer
from pygls.uris import from_fs_path

from yls import code_actions
from yls import icons
from yls import linting
from yls import utils
from yls.version import __version__
from yls.completer import Completer
from yls.hookspecs import ErrorMessage
from yls.hover import Hoverer
from yls.plugin_manager_provider import PluginManagerProvider
from yls.yaramod_provider import YaramodProvider

log = logging.getLogger(__name__)

utils.setup_logging()

# NOTE: All yaramod parsing can fail, create wrapper
# if we fail fallback to cached YaraFile (last valid)


class YaraLanguageServer(LanguageServer):
    """YaraLanguageServer implementation."""

    COMMAND_SCAN = "yls.scan"
    COMMAND_SCAN_ALL = "yls.scan_all"
    COMMAND_EVAL_SET_CONTEXT = "yls.eval_set_context"

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        log.debug("[__init__] LanguageServer constructor")

        # Yaramod parser provider
        self.ymod = YaramodProvider.instance()

        # Completion manager
        self.completer = Completer(self)

        # Hover manager
        self.hoverer = Hoverer(self)

        # Show debugging information on hover
        self.debug_hover = False

        # Lint the file when editing the document
        self.did_change_lint = False

        # Cached version of yara file used for completions
        # NOTE: We need to create dict not only one variable
        self.last_valid_yara_file = None

        self.register_plugin_commands()

    def register_plugin_commands(self) -> None:
        """Initialize the custom commands from plugins."""

        async def command_wrapper(command: str, ls: YaraLanguageServer, params: Any) -> Any:
            """Generic command hadler that delegates the custom commands to plugins."""
            log.info(f"Executing registered {command=} from plugins with {params=}")
            return await utils.pluggy_results(
                PluginManagerProvider.instance().hook.yls_execute_command(
                    command=command, ls=ls, params=params
                )
            )

        # Get the list of commands to register
        commands_to_register = utils.flatten_list(
            PluginManagerProvider.instance().hook.yls_supported_commands(ls=self)
        )
        for command in commands_to_register:
            func = functools.partial(command_wrapper, command)
            self.command(command)(func)

    def did_change_configuration(self) -> None:
        """Server changed the configuration.

        This function will fetch the latest configuration and set the YLS state accordingly.
        """

        def _config_callback(config: Any) -> None:
            try:
                conf = config[0]
                log.debug(f"[did_change_configuration] yls configuration is: {config[0]}")
                self.debug_hover = conf.debug.hover

            except Exception as e:  # pylint: disable=broad-except
                log.error(f"[did_change_configuration] Error occurred: {e}")

        # Request the configuration from the client for "yls"
        self.get_configuration(
            lsp_types.ConfigurationParams(
                items=[lsp_types.ConfigurationItem(scope_uri="", section="yls")]
            ),
            _config_callback,
        )

    def get_current_rule(
        self, uri: str, position: lsp_types.Position, yara_file: yaramod.YaraFile | None = None
    ) -> yaramod.Rule | None:
        """Get the last defined rule for given cursor position."""
        doc_path = pathlib.Path(self.workspace.get_document(uri).path)
        log.debug(f"[GET_CURRENT_RULE] Position={position}, Document={doc_path}")

        parsed_yfile = yara_file or utils.yaramod_parse_file(str(doc_path))
        if parsed_yfile is None:
            return None

        last_rule = None
        for rule in parsed_yfile.rules:
            if (
                rule.location.file_path
                and doc_path.exists()
                and doc_path.samefile(rule.location.file_path)
                and rule.location.begin.line > position.line
            ):
                return last_rule
            last_rule = rule
        return last_rule

    def get_references(
        self, uri: str, word: str, cursor: lsp_types.Position | None = None
    ) -> list[lsp_types.Location]:
        """Return list of locations that textually match given word."""
        text_doc = self.workspace.get_document(uri)
        source = text_doc.source
        res = []
        for m in re.finditer(re.escape(word), source):
            offset = m.start()
            line_counter = 0
            match_len = m.end() - m.start()
            for line in source.splitlines(keepends=True):
                line_len = len(line)
                if offset >= line_len:
                    offset -= line_len
                    line_counter += 1
                else:
                    # NOTE: Also we should check the character
                    if cursor and cursor.line == line_counter:
                        # Ignore this match
                        break

                    res.append(
                        lsp_types.Location(
                            uri=uri,
                            range=utils.range_from_coords(
                                (line_counter, offset), (line_counter, offset + match_len)
                            ),
                        )
                    )
                    break

        return res

    def show_error_message(self, msg: str, log_msg: str | None = None) -> None:
        """Show error message to user (also log it)."""
        log.error(f"[SHOW_ERROR_MESSAGE] {log_msg or msg}")
        self.show_message(msg, lsp_types.MessageType.Error)


# The main YLS Server instance
SERVER = YaraLanguageServer("yara-language-server", f"v{__version__}")


@SERVER.feature(WORKSPACE_DID_CHANGE_CONFIGURATION)
def did_change_configuration(ls: YaraLanguageServer, _params: Any) -> None:
    """Configuration was changed.

    This notification is somehow not being triggered. See `set_traceback` comment.
    """
    utils.log_command(WORKSPACE_DID_CHANGE_CONFIGURATION)
    log.debug("[DID_CHANGE_CONFIGURATION] Configuration was changed")
    ls.did_change_configuration()


@SERVER.feature("$/setTraceNotification")
def set_traceback(ls: YaraLanguageServer, params: Any) -> None:
    """Set traceback notification.

    This feature is declared manually because the pygls does not support it yet.
    It is proposed in the 3.16 version. Params include `TraceValue` which can be
    one of `{'off', 'message', 'verbose'}`.

    Based on tests it looks like this is send in the event of configuration change
    instead of WORKSPACE_DID_CHANGE_CONFIGURATION.
    """
    log.debug(
        "[SET_TRACE_NOTIFICATION] Received setTraceNotification, "
        "this could mean that the configuration was changed."
    )
    log.debug(params)
    ls.did_change_configuration()


@SERVER.feature(INITIALIZED)
def initiliazed(ls: YaraLanguageServer, _params: Any) -> None:
    """Connection is initialized."""
    utils.log_command(INITIALIZED)
    log.debug("[INITIALIZED] Connection was established")
    ls.did_change_configuration()
    # NOTE: In the future we can parse all files in the workspace


@SERVER.feature(COMPLETION, lsp_types.CompletionOptions(trigger_characters=["."]))
def completion(
    ls: YaraLanguageServer, params: lsp_types.CompletionParams
) -> lsp_types.CompletionList:
    """Code completion."""
    utils.log_command(COMPLETION)

    return ls.completer.complete(params)


@SERVER.feature(SIGNATURE_HELP, lsp_types.SignatureHelpOptions(trigger_characters=["("]))
def signature_help(
    ls: YaraLanguageServer, params: lsp_types.CompletionParams
) -> lsp_types.SignatureHelp | None:
    """Signature help."""
    utils.log_command(SIGNATURE_HELP)

    return ls.completer.signature_help(params)


@SERVER.feature(HOVER)
async def hover(
    ls: YaraLanguageServer, params: lsp_types.TextDocumentPositionParams
) -> lsp_types.Hover | None:
    """Cursor over information."""
    utils.log_command(HOVER)

    return await ls.hoverer.hover(params)


def lint(
    ls: YaraLanguageServer,
    params: lsp_types.DidOpenTextDocumentParams
    | lsp_types.DidChangeTextDocumentParams
    | lsp_types.DidSaveTextDocumentParams,
) -> None:
    """Lint and publish diagnostics."""
    # NOTE: DiagnosticRelatedInformation can be used to show the redefinition of a rule/string
    # Represents a related message and source code location for a diagnostic. This should be
    # used to point to code locations that cause or are related to a diagnostics, e.g when
    # duplicating a symbol in a scope.

    text_doc = ls.workspace.get_document(params.text_document.uri)
    log.info(f'[LINT] Lint request start for file "{text_doc.path}"')

    # Save the last valid yarafile, that can be used as a fallback in subsequent requests
    yara_file = utils.yaramod_parse_file(text_doc.path)
    if yara_file is not None:
        ls.last_valid_yara_file = yara_file

    diag = linting.get_diagnostics(text_doc)

    ls.publish_diagnostics(text_doc.uri, diag)


@SERVER.feature(TEXT_DOCUMENT_DID_OPEN)
def did_open(ls: YaraLanguageServer, params: lsp_types.DidOpenTextDocumentParams) -> None:
    """The document was opened."""
    utils.log_command(TEXT_DOCUMENT_DID_OPEN)
    lint(ls, params)


@SERVER.feature(TEXT_DOCUMENT_DID_SAVE)
def did_save(ls: YaraLanguageServer, params: lsp_types.DidSaveTextDocumentParams) -> None:
    """The document was saved."""
    utils.log_command(TEXT_DOCUMENT_DID_SAVE)
    lint(ls, params)


@SERVER.feature(TEXT_DOCUMENT_DID_CHANGE)
def did_change(ls: YaraLanguageServer, params: lsp_types.DidChangeTextDocumentParams) -> None:
    """The document was changed."""
    utils.log_command(TEXT_DOCUMENT_DID_CHANGE)
    if ls.did_change_lint:
        lint(ls, params)


@SERVER.feature(FORMATTING)
def formatting(
    ls: YaraLanguageServer, params: lsp_types.DocumentFormattingParams
) -> list[lsp_types.TextEdit]:
    """Format the whole buffer."""
    utils.log_command(FORMATTING)

    document = ls.workspace.get_document(params.text_document.uri)
    source = document.source
    path_str = document.path
    path = pathlib.Path(path_str)

    # Check if the file exists also on the filesystem, not only in the editor buffer.
    if not path.exists():
        ls.show_message("Please save the file before formatting.", lsp_types.MessageType.Warning)
        return []

    # Open current file from the disk using `binary` mode in order to preserve
    # original new-lines of the file.
    with open(path_str, "rb") as disk_yfile:
        disk_source = disk_yfile.read()

    # Compare contents with the editor buffer converted to bytes, in case they
    # are not the same the file was not saved to the disk and we should not proceed
    if source.encode() != disk_source:
        ls.show_message("Please save the file before formatting.", lsp_types.MessageType.Warning)
        return []

    res: list[lsp_types.TextEdit] = PluginManagerProvider.instance().hook.yls_formatting(
        ls=ls, params=params, document=document
    )

    return res


@SERVER.feature(DEFINITION)
def definition(
    ls: YaraLanguageServer, params: lsp_types.TextDocumentPositionParams
) -> list[lsp_types.Location] | None:
    """Jump to definition."""
    utils.log_command(DEFINITION)
    text_doc = ls.workspace.get_document(params.text_document.uri)
    token = utils.cursor_token(text_doc, params.position)
    if not token:
        return None

    res = []

    log.debug(f'[DEFINITION] Cursor in on token "{token}" with type "{token.type}"')
    if token.type == yaramod.TokenType.StringId:
        log.debug(f'[DEFINITION] Searching for "{token.text}" string definition')

        rule = ls.get_current_rule(params.text_document.uri, params.position)
        if rule is None:
            return None
        log.debug('[DEFINITION] Current rule "{rule}" with name "{rule.name}"')

        for string in rule.strings:
            if string.identifier == token.pure_text:
                res.append(
                    lsp_types.Location(
                        uri=params.text_document.uri,
                        range=utils.range_from_coords(
                            (string.location.begin.line - 1, string.location.begin.column - 1),
                            (string.location.end.line - 1, string.location.begin.column - 1),
                        ),
                    )
                )
                # Since strings cannot be redefined in a single rule there is
                # no reason to continue this loop and immediately return
                break
    elif token.type == yaramod.TokenType.Id:
        log.debug(f'[DEFINITION] Searching for "{token.text}" rule definition')
        path = ls.workspace.get_document(params.text_document.uri).path
        yara_file = utils.yaramod_parse_file(path)
        if yara_file is None:
            return []

        for rule in yara_file.rules:
            if rule.name == token.pure_text:
                res.append(
                    lsp_types.Location(
                        uri=from_fs_path(rule.location.file_path),
                        range=utils.range_from_coords(
                            (rule.location.begin.line - 1, rule.location.begin.column - 1),
                            (rule.location.end.line - 1, rule.location.end.column),
                        ),
                    )
                )

    return res


@SERVER.feature(REFERENCES)
def references(
    ls: YaraLanguageServer, params: lsp_types.ReferenceParams
) -> list[lsp_types.Location]:
    """Provide a list of references for the object under the cursor."""
    utils.log_command(REFERENCES)
    text_doc = ls.workspace.get_document(params.text_document.uri)
    cursor_string = utils.cursor_word(text_doc, params.position)
    log.debug(f'[REFERENCES] Searching for references of "{cursor_string}"')
    if cursor_string:
        return ls.get_references(params.text_document.uri, cursor_string, cursor=params.position)
    return []


@SERVER.feature(DOCUMENT_HIGHLIGHT)
def document_highlight(
    ls: YaraLanguageServer, params: lsp_types.TextDocumentPositionParams
) -> list[lsp_types.DocumentHighlight]:
    """Highlight references of the object under the cursor."""
    utils.log_command(DOCUMENT_HIGHLIGHT)
    text_doc = ls.workspace.get_document(params.text_document.uri)
    token = utils.cursor_token(text_doc, params.position)
    if not token:
        return []

    if token.type not in [
        yaramod.TokenType.Id,
        yaramod.TokenType.RuleName,
        yaramod.TokenType.StringId,
        yaramod.TokenType.StringIdAfterNewline,
    ]:
        return []

    refs = ls.get_references(params.text_document.uri, token.pure_text)
    res = [
        lsp_types.DocumentHighlight(
            range=utils.range_from_coords(
                (ref.range.start.line, ref.range.start.character),
                (ref.range.end.line, ref.range.end.character),
            )
        )
        for ref in refs
    ]

    return res


# NOTE: For now unimplemented
# @SERVER.feature(WORKSPACE_SYMBOL)
# def workspace_symbol(
#     ls: YaraLanguageServer, params: lsp_types.WorkspaceSymbolParams
# ) -> Optional[List[lsp_types.SymbolInformation]]:
#     """Create list of symbols in the current workspace."""
#     utils.log_command(WORKSPACE_SYMBOL)
#     # NOTE: This needs to find all *.yar files in the workspace and provide information on them
#     return None


@SERVER.feature(DOCUMENT_SYMBOL)
def document_symbol(
    ls: YaraLanguageServer, params: lsp_types.DocumentSymbolParams
) -> list[lsp_types.DocumentSymbol]:
    """Provide a list of symbols in the current document."""
    utils.log_command(DOCUMENT_SYMBOL)
    text_doc = ls.workspace.get_document(params.text_document.uri)

    res: list[lsp_types.DocumentSymbol] = []
    parsed_yfile = utils.yaramod_parse_file(text_doc.path)
    if parsed_yfile is None:
        return []

    for rule in utils.yaramod_rules_in_file(parsed_yfile):

        # Construct the string information for this rule. This is necessary because yls is
        # creating the whole document symbol hierarchy instead of flat list
        rule_strings: list[lsp_types.DocumentSymbol] = []
        for string in rule.strings:
            rule_strings.append(
                lsp_types.DocumentSymbol(
                    name=string.identifier,
                    detail=string.text,
                    kind=lsp_types.SymbolKind.String,
                    range=utils.range_from_coords(
                        (string.location.begin.line - 1, string.location.begin.column - 1),
                        (string.location.end.line, string.location.end.column),
                    ),
                    selection_range=utils.range_from_coords(
                        (string.location.begin.line - 1, string.location.begin.column - 1),
                        (string.location.end.line, string.location.end.column),
                    ),
                )
            )

        res.append(
            lsp_types.DocumentSymbol(
                name=rule.name,
                kind=lsp_types.SymbolKind.Class,
                range=utils.range_from_coords(
                    (rule.location.begin.line - 1, rule.location.begin.column),
                    (rule.location.end.line - 1, rule.location.end.column),
                ),
                selection_range=utils.range_from_coords(
                    (rule.location.begin.line - 1, rule.location.begin.column),
                    (rule.location.begin.line - 1, rule.location.begin.column),
                ),
                children=rule_strings,
            )
        )

    return res


@SERVER.feature(CODE_ACTION)
def code_action(
    ls: YaraLanguageServer, params: lsp_types.CodeActionParams
) -> list[lsp_types.Command | lsp_types.CodeAction] | None:
    """List available code actions for given context."""
    utils.log_command(CODE_ACTION)

    # Provide hoverer service with selected text range context
    ls.hoverer.selected_range = params.range
    return code_actions.from_params(ls, params)


def code_lens_eval(yara_file: yaramod.YaraFile) -> list[lsp_types.CodeLens]:
    """Create evaluation code lenses from YaraFile."""
    yls_eval_enabled = any(PluginManagerProvider.instance().hook.yls_eval_enabled())
    if not yls_eval_enabled:
        return []

    res = []
    for rule in utils.yaramod_rules_in_file(yara_file):
        for meta in rule.metas:
            # Consider only hash metas that have valid hash as value
            if meta.key != "hash" and not utils.is_sha256_hash(meta.value.pure_text):
                continue

            # Create code lens for debugging
            lens = lsp_types.CodeLens(
                range=utils.range_from_coords(
                    (meta.token_key.location.begin.line - 1, meta.token_key.location.begin.column),
                    (meta.token_value.location.end.line - 1, meta.token_value.location.end.column),
                ),
                command=lsp_types.Command(
                    title=f"{icons.SEARCH} Select hash for context",
                    command=YaraLanguageServer.COMMAND_EVAL_SET_CONTEXT,
                    arguments=[
                        meta.value.pure_text,
                        utils.extract_rule_context_from_yarafile(yara_file, rule),
                    ],
                ),
            )
            res.append(lens)

    return res


def code_lens_scan(yara_file: yaramod.YaraFile, uri: str) -> list[lsp_types.CodeLens]:
    """Create scannig code lenses from YaraFile."""
    yls_scan_enabled = any(PluginManagerProvider.instance().hook.yls_scan_enabled())
    if not yls_scan_enabled:
        return []

    res = []
    for rule in utils.yaramod_rules_in_file(yara_file):
        if not utils.yaramod_rule_has_hashes(rule):
            continue

        # Create code lens for scans
        lens_scan = lsp_types.CodeLens(
            range=utils.range_from_yaramod_token(rule.token_first),
            command=lsp_types.Command(
                title=f"{icons.PLAY} Scan",
                command=YaraLanguageServer.COMMAND_SCAN,
                arguments=[uri, rule.name],
            ),
        )
        res.append(lens_scan)

    return res


@SERVER.feature(CODE_LENS)
def code_lens(ls: YaraLanguageServer, params: lsp_types.CodeLensParams) -> list[lsp_types.CodeLens]:
    utils.log_command(CODE_LENS)

    text_doc = ls.workspace.get_document(params.text_document.uri)
    path = text_doc.path

    yara_file = utils.yaramod_parse_file(path)
    if yara_file is None:
        return []

    res = []
    res.extend(
        utils.flatten_list(
            PluginManagerProvider.instance().hook.yls_code_lens(ls=ls, params=params)
        )
    )

    res.extend(code_lens_eval(yara_file))
    res.extend(code_lens_scan(yara_file, params.text_document.uri))

    return res


@SERVER.command(YaraLanguageServer.COMMAND_SCAN)
async def command_scan(ls: YaraLanguageServer, args: list[Any]) -> None:
    utils.log_command(YaraLanguageServer.COMMAND_SCAN)
    log.debug(f"{args=}")

    if len(args) != 2:
        return

    file_uri = args[0]
    rule_name = args[1]
    document = ls.workspace.get_document(file_uri)

    scan_results = await utils.pluggy_results(
        PluginManagerProvider.instance().hook.yls_scan(
            ls=ls, document=document, rule_name=rule_name
        )
    )

    for diagnostics in scan_results:
        ls.publish_diagnostics(document.uri, diagnostics)


@SERVER.command(YaraLanguageServer.COMMAND_SCAN_ALL)
async def command_scan_all(ls: YaraLanguageServer, args: list[Any]) -> None:
    utils.log_command(YaraLanguageServer.COMMAND_SCAN_ALL)
    log.debug(f"{args=}")

    if len(args) != 1:
        return
    file_uri = args[0]

    document = ls.workspace.get_document(file_uri)

    scan_results = await utils.pluggy_results(
        PluginManagerProvider.instance().hook.yls_scan(ls=ls, document=document, rule_name=None)
    )

    for diagnostics in scan_results:
        ls.publish_diagnostics(document.uri, diagnostics)


@SERVER.command(YaraLanguageServer.COMMAND_EVAL_SET_CONTEXT)
async def command_eval_set_context(ls: YaraLanguageServer, args: list[Any]) -> None:
    utils.log_command(YaraLanguageServer.COMMAND_EVAL_SET_CONTEXT)
    log.debug(f"{args=}")

    if len(args) != 2 or not utils.is_sha256_hash(args[0]):
        return

    _hash = args[0]
    ruleset = args[1]

    res_set_contexts = await utils.pluggy_results(
        PluginManagerProvider.instance().hook.yls_eval_set_context(
            ls=ls, _hash=_hash, ruleset=ruleset
        )
    )

    res_success = next((res for res in res_set_contexts if not isinstance(res, ErrorMessage)), None)

    if res_success:
        res_success.show(ls)
    else:
        for i, res_set_context in enumerate(res_set_contexts):
            if len(res_set_contexts) > 1:
                res_set_context.message = f"{utils.DEBUGGER_SOURCES[i]}: {res_set_context.message}"
            log.debug(res_set_context.message)
            res_set_context.show(ls)


def main() -> None:
    """YLS entry point."""

    # Parse command line arguments
    args = utils.create_options_parser().parse_args()

    utils.set_logging_level(args.verbose)
    utils.logging_prolog(PluginManagerProvider.instance())

    SERVER.start_io()


if __name__ == "__main__":
    main()



from __future__ import annotations

import re
from dataclasses import dataclass
from enum import Enum
from functools import lru_cache

import pygls.lsp.types as lsp_types

from yls import utils


class StringType(Enum):
    """Type of the YARA string."""

    PLAIN = "text"
    REGEX = "regexp"
    HEX = "hex"


@dataclass
class StringModifier:
    keyword: str
    string_types: list[StringType]
    summary: str
    restrictions: str

    def to_completion_item(self) -> lsp_types.CompletionItem:
        return lsp_types.CompletionItem(
            label=self.keyword,
            kind=lsp_types.CompletionItemKind.Keyword,
            insert_text=self.keyword,
            insert_text_format=lsp_types.InsertTextFormat.PlainText,
            documentation=utils.markdown_content(self.documentation),
            sort_text="a",  # High priority
        )

    @property
    def documentation(self) -> str:
        return f"""`{self.keyword}`

**Summary:** {self.summary}

**String types:** {", ".join(map(lambda s: s.value, self.string_types))}

**Restrictions:** {self.restrictions}
"""


STRING_MODIFIERS = {
    "nocase": StringModifier(
        "nocase",
        [StringType.PLAIN, StringType.REGEX],
        "Ignore case",
        "Cannot use with `xor`, `base64`, or `base64wide`",
    ),
    "wide": StringModifier(
        "wide",
        [StringType.PLAIN, StringType.REGEX],
        "Emulate UTF16 by interleaving null (0x00) characters",
        "None",
    ),
    "ascii": StringModifier(
        "ascii",
        [StringType.PLAIN, StringType.REGEX],
        "Also match ASCII characters, only required if `wide` is used",
        "None",
    ),
    "xor": StringModifier(
        "xor",
        [StringType.PLAIN],
        "XOR text string with single byte keys",
        "Cannot use with `nocase`, `base64`, or `base64wide`",
    ),
    "base64": StringModifier(
        "base64",
        [StringType.PLAIN],
        "Convert to 3 base64 encoded strings",
        "Cannot use with `nocase`, `xor`, or `fullword`",
    ),
    "base64wide": StringModifier(
        "base64wide",
        [StringType.PLAIN],
        "Convert to 3 base64 encoded strings, then interleaving null characters like wide",
        "Cannot use with `nocase`, `xor`, or `fullword`",
    ),
    "fullword": StringModifier(
        "fullword",
        [StringType.PLAIN, StringType.REGEX],
        "Match is not preceded or followed by an alphanumeric character",
        "Cannot use with `base64`, or `base64wide`",
    ),
    "private": StringModifier(
        "private",
        [StringType.PLAIN, StringType.REGEX, StringType.HEX],
        "Match never included in output",
        "None",
    ),
}


@lru_cache
def string_modifiers_completion_items(
    string_type: StringType | None,
) -> list[lsp_types.CompletionItem]:
    res = []
    for string_modifier in STRING_MODIFIERS.values():
        if string_type not in string_modifier.string_types:
            continue

        res.append(string_modifier.to_completion_item())

    return res


def estimate_string_type(line: str) -> StringType | None:
    """Estimate YARA string type based on the line from the source.

    This is only an estimation, since we have to also consider unfinished
    strings. We can pretty reliably estimate this based on the opening
    character after the equals sign. If the estimation cannot be done, we
    return `None`.
    """
    if re.search(r'=\s*"', line) is not None:
        return StringType.PLAIN
    elif re.search(r"=\s*/", line) is not None:
        return StringType.REGEX
    elif re.search(r"=\s*{", line) is not None:
        return StringType.HEX
    else:
        return None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['yls']

package_data = \
{'': ['*']}

install_requires = \
['pluggy>=1.0.0,<2.0.0',
 'pygls==0.13.0',
 'yaramod>=3.16.0,<4.0.0',
 'yari-py>=0.1.1,<0.2.0']

entry_points = \
{'console_scripts': ['yls = yls.server:main']}

setup_kwargs = {
    'name': 'yls',
    'version': '1.2.2',
    'description': 'YARA Language Server',
    'long_description': '# YLS\n\n![PyPI](https://img.shields.io/pypi/v/yls?label=yls)\n![Visual Studio Marketplace Version](https://img.shields.io/visual-studio-marketplace/v/avast-threatlabs-yara.vscode-yls?label=vscode)\n\n[Language server](https://microsoft.github.io/language-server-protocol/) for\n[YARA](https://yara.readthedocs.io/en/stable/) language.\n\n:rocket: Features:\n- Code completion of all available modules (including function parameters)\n- Function documentation for hovers and code completion\n- Opinionated code formatting\n- Signature help\n- Linting\n- Go-to definition and references\n- Symbol highlighting under the cursor\n- Debugging? Stay tuned...\n- ...\n\n![Showcase](https://github.com/avast/yls/raw/master/docs/assets/yls.png)\n\nFor more information, check out:\n- [Blog post](https://engineering.avast.io/yls-first-step-towards-yara-development-environment/)\n- [Wiki](https://www.github.com/avast/yls/wiki)\n\n:snake: Minimal supported version of Python is `3.8`.\n\n## Installation\n\nTo setup your environment please follow instructions on\n[wiki](https://github.com/avast/yls/wiki/How-to-setup).\n\n## How to develop\n\nInstall YLS in development mode with all necessary dependencies.\n\n```bash\npoetry install\n```\n\n### Tests\n\nYou can run tests with the following command:\n\n```bash\npoetry run pytest\n```\n\n## License\n\nCopyright (c) 2022 Avast Software, licensed under the MIT license. See the\n[`LICENSE`](https://github.com/avast/yls/blob/master/LICENSE) file for more\ndetails.\n\nYLS and its related projects uses third-party libraries or other resources\nlisted, along with their licenses, in the\n[`LICENSE-THIRD-PARTY`](https://github.com/avast/yls/blob/master/LICENSE-THIRD-PARTY)\nfile.\n\n## FAQ\n\n### Why are you using `pluggy`?\n\nSome parts depend on our internal services, however we are working on making\nmost of the code available. This is just the first piece.\n',
    'author': 'Matej Kašťák',
    'author_email': 'matej.kastak@avast.com',
    'maintainer': 'Matej Kašťák',
    'maintainer_email': 'matej.kastak@avast.com',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.12',
}


setup(**setup_kwargs)



from __future__ import annotations

import argparse
import asyncio
import itertools
import logging
import os
import pathlib
import platform
import re
import sys
import uuid
from typing import Any
from typing import Iterator
from typing import TypeVar

import pygls.lsp.types as lsp_types
import yaramod
from pluggy import PluginManager
from pygls.workspace import Document

from yls.version import __version__
from yls.yaramod_provider import YaramodProvider

log = logging.getLogger(__name__)


DEBUGGER_SOURCES = ["avast", "core"]


def create_options_parser() -> argparse.ArgumentParser:
    """Create argparse parser object."""
    parser = argparse.ArgumentParser(description="Yara Language Server")
    parser.add_argument("-v", "--verbose", action="count", help="verbosity level", default=0)
    parser.add_argument("--version", action="version", version=f"%(prog)s {__version__}")
    return parser


def setup_logging() -> None:
    """Setup the default application logging."""
    # Configure the application logging
    root_logger = logging.getLogger()

    # Remove old handlers
    for handler in root_logger.handlers:
        root_logger.removeHandler(handler)
        handler.close()

    log_format = "[{asctime}] [{name:>15}:{lineno:<3}] [{levelname:.4}] -- {message}"

    # Default logging is to stderr
    steam_logging_handler = logging.StreamHandler(stream=sys.stderr)
    steam_logging_handler.setFormatter(logging.Formatter(log_format, style="{"))
    root_logger.addHandler(steam_logging_handler)

    yls_log_file = os.environ.get("YLS_LOG_FILE")
    if yls_log_file:
        file_logging_handler = logging.FileHandler(yls_log_file)
        file_logging_handler.setFormatter(logging.Formatter(log_format, style="{"))
        root_logger.addHandler(file_logging_handler)


def set_logging_level(verbose_count: int) -> None:
    """Set the logging level based on the number of -v options on command line."""
    log_levels = [logging.WARNING, logging.INFO, logging.DEBUG]
    log_level = log_levels[min(verbose_count, len(log_levels) - 1)]
    root_logger = logging.getLogger()
    root_logger.setLevel(log_level)


def log_command(command: str) -> None:
    """Log command in a standard format.

    NOTE: Be careful when changing this format, we estimate metrics from it.
    """
    if not command:
        return

    log.info(f"LSP_COMMAND = {command}")


def logging_prolog(plugin_manager: PluginManager) -> None:
    """Emit dependency version information."""
    log.info("Starting yls language server...")
    log.info(f"System platform: {platform.system()}")
    log.info(f"YLS version: {__version__}")
    log.info(f"Yaramod version: {yaramod.YARAMOD_VERSION}")

    plugins = ", ".join(plugin[0] for plugin in plugin_manager.list_name_plugin())
    log.info(f"Activated plugins: {plugins}")


def range_from_coords(x: tuple[int, int], y: tuple[int, int]) -> lsp_types.Range:
    """Helper function to create a range from coordinates.

    NOTE: If any of the coordinates is negative, truncate it to 0."""
    return lsp_types.Range(
        start=lsp_types.Position(line=max(0, x[0]), character=max(0, x[1])),
        end=lsp_types.Position(line=max(0, y[0]), character=max(0, y[1])),
    )


def range_from_yaramod_location(location: yaramod.Location) -> lsp_types.Range:
    """Helper function to create a range from yaramod location."""
    return range_from_coords(
        (location.begin.line - 1, location.begin.column - 1),
        (location.end.line - 1, location.end.column),
    )


def range_from_yaramod_token(token: yaramod.Token) -> lsp_types.Range:
    """Helper function to create a range from yaramod token."""
    return range_from_yaramod_location(token.location)


def range_from_yaramod_rule(rule: yaramod.Rule) -> lsp_types.Range:
    """Helper function to create a range from yaramod rule."""
    return range_from_coords(
        (rule.location.begin.line - 1, rule.location.begin.column - 1),
        (rule.location.end.line - 1, rule.location.end.column),
    )


def range_from_yaramod_expression(expr: yaramod.Expression) -> lsp_types.Range:
    """Convert Yaramod expression to LSP range."""
    end = lsp_types.Position(
        line=expr.token_last.location.end.line - 1, character=expr.token_last.location.end.column
    )
    start = lsp_types.Position(
        line=expr.token_first.location.begin.line - 1,
        character=expr.token_first.location.begin.column - 1,
    )
    return lsp_types.Range(start=start, end=end)


def range_from_line(
    line: int, document: Document, strip_whitespace: bool = True
) -> lsp_types.Range:
    """Helper function to create a range from line index.

    This function will ignore any leading/trailing whitespace."""
    line_content = document.lines[line]
    line_len = len(line_content)

    if strip_whitespace:
        start_char = line_len - len(line_content.lstrip())
        end_char = len(line_content.rstrip())
    else:
        start_char = 0
        end_char = line_len

    return range_from_coords((line, start_char), (line, end_char))


def range_from_lines(
    first_line: int, last_line: int, document: Document, strip_whitespace: bool = True
) -> lsp_types.Range:
    """Helper function to create a range for multiple lines.

    This function will ignore any leading/trailing whitespace."""
    first = range_from_line(first_line, document, strip_whitespace)
    last = range_from_line(last_line, document, strip_whitespace)
    return range_from_coords(
        (first.start.line, first.start.character), (last.end.line, last.end.character)
    )


def range_for_first_line() -> lsp_types.Range:
    """Create range for the first line.

    Editors should bound the range to an actual line length `min(len(line), 200)`."""
    return range_from_coords((0, 0), (0, 200))


T = TypeVar("T", str, bytes)


def remove_prefix(source: T, prefix: T) -> T:
    """Remove prefix from string or bytes if present."""
    if source.startswith(prefix):
        return source[len(prefix) :]
    return source


def remove_suffix(source: T, suffix: T) -> T:
    """Remove suffix from string or bytes if present."""
    if source.endswith(suffix):
        return source[: len(source) - len(suffix)]
    return source


def first_non_space_pos(line: str) -> int:
    """Return the index of first non space character in string.

    This is useful when calculating the `Range` and the leading spaces should
    not be included in this range."""
    counter = 0
    for c in line:
        if not c.isspace():
            return counter
        counter += 1
    return 0


def document_line(document: Document, line: int, keepends: bool = False) -> str:
    """Return the line of the document."""
    try:
        line_str = document.source.splitlines(keepends)[line]
    except IndexError:
        return ""
    return line_str


def cursor_line(document: Document, position: lsp_types.Position, keepends: bool = False) -> str:
    """Return the line the cursor is on."""
    return document_line(document, position.line, keepends=keepends)


def cursor_symbol(document: Document, position: lsp_types.Position) -> str | None:
    """Return the symbol under the cursor."""
    line = cursor_line(document, position)
    cursor = position.character
    for m in re.finditer(r"[\w$*.]+", line):
        if m.start() <= cursor <= m.end():
            return line[m.start() : m.end()]
    return None


def cursor_word(
    document: Document, position: lsp_types.Position, include_all: bool = True
) -> str | None:
    """Return the word under the cursor."""
    res = cursor_word_and_range(document, position, include_all=include_all)
    if res:
        return res[0]
    return None


def cursor_word_and_range(
    document: Document, position: lsp_types.Position, include_all: bool = True
) -> tuple[str, lsp_types.Range] | None:
    """Return the word and its range under the cursor."""
    line = cursor_line(document, position)
    cursor = position.character
    for m in re.finditer(r"[\w$*.()\/\\#:]+", line):
        end = m.end() if include_all else cursor
        if m.start() <= cursor <= m.end():
            word = (
                line[m.start() : end],
                range_from_coords((position.line, m.start()), (position.line, end)),
            )
            return word
    return None


def cursor_token(document: Document, position: lsp_types.Position) -> yaramod.Token | None:
    """Return the token under the cursor."""
    yara_file = yaramod_parse_file(document.path)
    if yara_file is None:
        return None

    # Iterate over tokens and find token with matching position
    for token in yara_file.tokenstream.tokens:
        if (
            token.location.file_path
            and os.path.samefile(token.location.file_path, document.path)
            and token.location.begin.line == position.line + 1
            and token.location.begin.column <= position.character + 1
            and token.location.end.column >= position.character + 1
        ):
            return token

    return None


def is_in_yara_section(document: Document, cursor_line: int, section_name: str) -> bool:
    found_beginning = False
    found_end = False

    if len(document.lines) <= cursor_line:
        return False

    # Walk back from cursor position (cursor_line included)
    for i in reversed(range(0, cursor_line + 1)):
        if re.match(r"^}\s*$", document.lines[i]):
            # Found the end of previous rule, which means cursor is not pointing into a rule
            return False

        if re.match(rf"^(({{)|(rule \w+)|((?!\s+{section_name}:)\s+[a-z]+:))", document.lines[i]):
            # Found the beginning of current rule or different section, stop searching for section_name
            break

        if re.match(rf"\s+{section_name}:", document.lines[i]):
            found_beginning = True
            break

    if found_beginning:
        # Walk forward from cursor position
        for i in range(cursor_line + 1, len(document.lines)):
            if re.match(r"^((\s+[a-z]+:)|(}))\s*$", document.lines[i]):
                # Found the next section or the end of current rule
                found_end = True
                break
    return found_beginning and found_end


def position_in_range(position: lsp_types.Position, _range: lsp_types.Range) -> bool:
    """Returns if a given position is contained within the specified range."""
    return (
        _range.start.line < position.line
        or (_range.start.line == position.line and _range.start.character <= position.character)
    ) and (
        _range.end.line > position.line
        or (_range.end.line == position.line and _range.end.character >= position.character)
    )


# pylint: disable-next=too-many-public-methods
class YaramodExpressionExtractor(yaramod.ObservingVisitor):  # type: ignore
    """Helper to extract interesting expressions."""

    def __init__(self, position: lsp_types.Position) -> None:
        super().__init__()
        self.position = position
        self.expr = None

    def run(self, condition: yaramod.Expression) -> yaramod.Expression | None:
        self.observe(condition)
        return self.expr

    def extract(self, expr: yaramod.Expression) -> None:
        expr_range = range_from_yaramod_expression(expr)
        if position_in_range(self.position, expr_range):
            self.expr = expr

    # pylint: disable-next=invalid-name
    def visit_FunctionCallExpression(self, expr: yaramod.FunctionCallExpression) -> None:
        self.extract(expr)

    # pylint: disable-next=invalid-name
    def visit_IdExpression(self, expr: yaramod.IdExpression) -> None:
        self.extract(expr)

    # pylint: disable-next=invalid-name
    def visit_StructAccessExpression(self, expr: yaramod.StructAccessExperssion) -> None:
        self.extract(expr)

    # pylint: disable-next=invalid-name
    def visit_ArrayAccessExpression(self, expr: yaramod.ArrayAccessExpression) -> None:
        self.extract(expr)

    # pylint: disable-next=invalid-name
    def visit_StringExpression(self, expr: yaramod.StringExpression) -> None:
        self.extract(expr)

    # pylint: disable-next=invalid-name
    def visit_StringCountExpression(self, expr: yaramod.StringCountExpression) -> None:
        self.extract(expr)

    # pylint: disable-next=invalid-name
    def visit_StringOffsetExpression(self, expr: yaramod.StringOffsetExpression) -> None:
        self.extract(expr)

    # pylint: disable-next=invalid-name
    def visit_StringLengthExpression(self, expr: yaramod.StringLengthExpression) -> None:
        self.extract(expr)

    # pylint: disable-next=invalid-name
    def visit_OfExpression(self, expr: yaramod.OfExpression) -> None:
        self.extract(expr)

    # pylint: disable-next=invalid-name
    def visit_ForExpression(self, expr: yaramod.ForExpression) -> None:
        self.extract(expr)

    # pylint: disable-next=invalid-name
    def visit_IequalsExpression(self, expr: yaramod.IequalsExpression) -> None:
        self.extract(expr)
        expr.left_operand.accept(self)
        expr.right_operand.accept(self)

    # pylint: disable-next=invalid-name
    def visit_LtExpression(self, expr: yaramod.LtExpression) -> None:
        self.extract(expr)
        expr.left_operand.accept(self)
        expr.right_operand.accept(self)

    # pylint: disable-next=invalid-name
    def visit_GtExpression(self, expr: yaramod.GtExpression) -> None:
        self.extract(expr)
        expr.left_operand.accept(self)
        expr.right_operand.accept(self)

    # pylint: disable-next=invalid-name
    def visit_LeExpression(self, expr: yaramod.LeExpression) -> None:
        self.extract(expr)
        expr.left_operand.accept(self)
        expr.right_operand.accept(self)

    # pylint: disable-next=invalid-name
    def visit_GeExpression(self, expr: yaramod.GeExpression) -> None:
        self.extract(expr)
        expr.left_operand.accept(self)
        expr.right_operand.accept(self)

    # pylint: disable-next=invalid-name
    def visit_EqExpression(self, expr: yaramod.EqExpression) -> None:
        self.extract(expr)
        expr.left_operand.accept(self)
        expr.right_operand.accept(self)

    # pylint: disable-next=invalid-name
    def visit_NeqExpression(self, expr: yaramod.NeqExpression) -> None:
        self.extract(expr)
        expr.left_operand.accept(self)
        expr.right_operand.accept(self)

    # pylint: disable-next=invalid-name
    def visit_AndExpression(self, expr: yaramod.AndExpression) -> None:
        self.extract(expr)
        expr.left_operand.accept(self)
        expr.right_operand.accept(self)

    # pylint: disable-next=invalid-name
    def visit_OrExpression(self, expr: yaramod.OrExpression) -> None:
        self.extract(expr)
        expr.left_operand.accept(self)
        expr.right_operand.accept(self)

    # pylint: disable-next=invalid-name
    def visit_NotExpression(self, expr: yaramod.NotExpression) -> None:
        self.extract(expr)
        expr.operand.accept(self)

    # pylint: disable-next=invalid-name
    def visit_DefinedExpression(self, expr: yaramod.DefinedExpression) -> None:
        self.extract(expr)
        expr.operand.accept(self)

    # pylint: disable-next=invalid-name
    def visit_ContainsExpression(self, expr: yaramod.ContainsExpression) -> None:
        self.extract(expr)
        expr.left_operand.accept(self)
        expr.right_operand.accept(self)

    # pylint: disable-next=invalid-name
    def visit_MatchesExpression(self, expr: yaramod.MatchesExpression) -> None:
        self.extract(expr)
        expr.left_operand.accept(self)
        expr.right_operand.accept(self)


def cursor_expression(
    yfile: yaramod.YaraFile, position: lsp_types.Position
) -> yaramod.Expression | None:
    """Return the expression under the cursor."""

    extractor = YaramodExpressionExtractor(position)
    for rule in yaramod_rules_in_file(yfile):
        expr = extractor.run(rule.condition)
        if expr is not None:
            return expr

    return None


def debug_hover(document: Document, position: lsp_types.Position) -> lsp_types.Hover:
    token = cursor_token(document, position)
    if token:
        return lsp_types.Hover(
            contents=markdown_content(
                f"{token.text=}\n\n"  # Double newline to force line breaks
                f"{token.pure_text=}\n\n"
                f"{token.type=}\n\n"
                f"Begin = ({token.location.begin.line}, {token.location.begin.column})\n\n"
                f"End = ({token.location.end.line}, {token.location.end.column})\n\n"
                f"LSP Position = {position}\n\n"
                f"_cursor_word = {cursor_word(document, position)}"
            )
        )
    return lsp_types.Hover(contents=f"Token = {token}\nLSP Position = {position}")


def is_sha256_hash(i: str) -> bool:
    """Check if the `i` looks like a sha256 hash.

    We accept only lowercase hashes because it is the same behavior as YRTC."""
    return re.fullmatch("[0-9a-f]{64}", i) is not None


def markdown_content(value: str) -> lsp_types.MarkupContent:
    """Return the MarkupContent with Markdown kind."""
    return lsp_types.MarkupContent(kind=lsp_types.MarkupKind.Markdown, value=value)


def yaramod_parse_file(path: str) -> yaramod.YaraFile | None:
    """Wrapper around `Yaramod().parse_file()`."""
    try:
        return YaramodProvider.instance().parse_file(path)
    except yaramod.ParserError as e:
        log.debug(f'[YARAMOD_PARSE_FILE] Yaramod parse exception: "{e}"')
        return None


def yaramod_rules_in_file(yfile: yaramod.YaraFile) -> Iterator[yaramod.Rule]:
    """Yield only rules from the file specified by path.

    YaraFile contains all rules (even ones that come from includes). Most of the times we care
    only about the rules that are coming the source code file.
    """
    path = yarafile_path(yfile)
    for rule in yfile.rules:
        # Ignore rules coming from include statements
        if path is not None and not pathlib.Path(rule.location.file_path).samefile(
            pathlib.Path(path)
        ):
            continue

        yield rule


def yaramod_rule_has_hashes(rule: yaramod.Rule) -> bool:
    """Check if YARA rule contains sample hashes in hash meta."""
    for meta in rule.metas:
        # Consider only hash metas that have valid hash as value
        if meta.key != "hash" or not is_sha256_hash(meta.value.pure_text):
            continue

        return True

    return False


def yaramod_expression_to_symbol(expr: yaramod.StructAccessExpression) -> str | None:
    if isinstance(expr, yaramod.StructAccessExpression):
        base_expr = expr
    elif isinstance(expr, yaramod.FunctionCallExpression):
        base_expr = expr.function
    elif isinstance(expr, yaramod.ArrayAccessExpression):
        # For some reason array access has the full symbol path in expr.symbol.name
        # This is not the case for other kinds of expressions
        return f"{expr.symbol.name}[]"
    elif isinstance(expr, yaramod.FilesizeExpression):
        return "filesize"
    else:
        log.warning(f"Failed to convert {expr=} to symbol string.")
        return None

    cursor = base_expr
    func_name = ""
    while True:
        if isinstance(cursor, yaramod.StructAccessExpression):
            func_name = f"{cursor.symbol.name}.{func_name}"
            cursor = cursor.structure
        elif isinstance(cursor, yaramod.ArrayAccessExpression):
            func_name = f"{cursor.symbol.name}[].{func_name}"
            cursor = cursor.array.structure
        elif isinstance(cursor, yaramod.IdExpression):
            func_name = f"{cursor.symbol.name}.{func_name}"
            break
        else:
            log.warning(
                f"Error while converting expression to function name, unknown type {cursor=}"
            )
            return None

    return func_name[:-1]


def yarafile_to_string(yara_file: yaramod.YaraFile) -> str:
    """Explode YaraFile (and all includes) to a single string.

    This function will try import all modules.
    """

    ruleset = """import "cuckoo"
import "elf"
import "hash"
import "math"
import "pe"

"""
    for rule in yara_file.rules:
        ruleset += rule.text
        ruleset += "\n"

    return ruleset


def yarafile_path(yfile: yaramod.YaraFile) -> str | None:
    """Estimate the path of YaraFile.

    We are trying to ignore rules that come from `include`. Rules should be in order they
    appear in the source code. Getting the last rule should ensure, we get the rule that is
    from the source code even if there are included rules.

    If the file does not exist or YaraFile is created by parsing a string return None. This
    is also the case when there are no rules in the YaraFile.
    """
    rules = yfile.rules
    if len(rules) == 0:
        return None

    path: str = rules[-1].location.file_path
    if path == "[stream]":
        return None

    if not pathlib.Path(path).exists():
        return None

    return path


def yarafile_get_rule(
    yara_file: yaramod.YaraFile, rule_name: str, ignore_includes: bool = True
) -> yaramod.Rule | None:
    """Search for rule in YaraFile with a given name.

    If `ignore_includes` is True, consider only rules that are not from includes.
    """
    rules = yaramod_rules_in_file(yara_file) if ignore_includes else yara_file.rules
    for rule in rules:
        if rule.name == rule_name:
            return rule
    return None


def generate_progress_token() -> str:
    """Generate unique token that can be used to signalize progress."""
    return str(uuid.uuid4())


def flatten_list(list_of_lists: list[list[Any]]) -> list[Any]:
    return list(itertools.chain(*list_of_lists))


async def pluggy_results(objs: list[Any]) -> Any:
    res = []
    for obj in objs:
        resolved_object = await pluggy_result(obj)
        res.append(resolved_object)
    return res


async def pluggy_result(obj: Any) -> Any:
    """Get the result from YLS plugin. Try to extract the value from object.

    Taken from: https://simonwillison.net/2020/Sep/2/await-me-maybe/
    """
    log.debug(f"Resolving {obj=}")

    if callable(obj):
        return obj()
    if asyncio.iscoroutine(obj):
        return await obj
    return obj


async def start_progress(ls: Any, msg: str) -> str | None:
    """Helper to kick-start the progress indicator in editor."""
    token = generate_progress_token()
    log.debug(f"Generated a new {token=}")

    try:
        await ls.progress.create_async(token)
    except Exception:  # pylint: disable=broad-except
        # For some reason they raise plain exception when the token is already used
        ls.show_error_message("Progress token is already used")
        return None

    if msg:
        ls.progress.begin(
            token, lsp_types.WorkDoneProgressBegin(kind="begin", title=msg, percentage=0)
        )

    return token


class RuleDependencyIdentifier(yaramod.ObservingVisitor):  # type: ignore
    """A class for obtaining rules that are in the condition (recursively)."""

    def __init__(self) -> None:
        super().__init__()
        self.discovered_rules: set[str] = set()
        self.last_rule: yaramod.Rule = None

    def indirectly_affected(self, yara_file: yaramod.YaraFile, root_rule: str) -> set[str]:
        self.discovered_rules = {root_rule}

        last_discovered: set[str] = set()
        while last_discovered != self.discovered_rules:
            last_discovered = set(self.discovered_rules)
            for rule in yara_file.rules:
                self.last_rule = rule
                self.observe(rule.condition)

        return self.discovered_rules

    def visit_IdExpression(self, expr: yaramod.IdExpression) -> None:
        if self.last_rule and self.last_rule.name in self.discovered_rules:
            self.discovered_rules.add(expr.symbol.name)


def extract_rule_context_from_yarafile(yara_file: yaramod.YaraFile, rule: yaramod.Rule) -> str:
    """Extract specified rule, private rules and modules it is dependent on from YaraFile."""

    # Get rules ordered from deepest dependency to `rule` so they can be compiled.
    rule_names = RuleDependencyIdentifier().indirectly_affected(yara_file, rule.name)
    rule_context = "\n\n".join(rule.text for rule in yara_file.rules if rule.name in rule_names)

    # Import only modules that are needed by context
    imports = "\n".join(
        f'import "{module.name}"'
        for module in yara_file.imports
        if re.search(rf"\W{module.name}\.", rule_context)
    )
    if imports:
        rule_context = imports + "\n\n" + rule_context

    return rule_context


def text_from_range(document: Document, expr_range: lsp_types.Range) -> str:
    """Extract text from `document` which is located in provided `range`"""
    first_line = document_line(document, expr_range.start.line)

    if expr_range.start.line == expr_range.end.line:
        return first_line[expr_range.start.character : expr_range.end.character]
    else:
        expr = first_line[expr_range.start.character :]

        for line_idx in range(expr_range.start.line + 1, expr_range.end.line):
            expr += document_line(document, line_idx) + "\n"

        end_line = document_line(document, expr_range.end.line)
        expr += end_line[0 : expr_range.end.character + 1]
        return expr


def range_to_expression(document: Document, _range: lsp_types.Range) -> lsp_types.Range | None:
    """Converts selected range in expression to the smallest possible Yaramod expression that covers the entire range."""
    yara_file = yaramod_parse_file(document.path)
    if yara_file is None:
        return None

    start = cursor_expression(yara_file, _range.start)
    end = cursor_expression(yara_file, _range.end)

    if start and end:
        return lsp_types.Range(
            start=range_from_yaramod_expression(start).start,
            end=range_from_yaramod_expression(end).end,
        )
    else:
        return None


def remove_whitespace(text: str) -> str:
    """Remove redundant whitespaces to display text in a single line."""
    return re.sub(r"\s+", " ", text)


def truncate_message(text: str, limit: int = 150) -> str:
    return (text[:limit] + "..") if len(text) > limit else text


async def get_config_from_editor(ls: Any, section: str) -> Any:
    """Get the configuration from the editor."""
    config = await ls.get_configuration_async(
        lsp_types.ConfigurationParams(
            items=[lsp_types.ConfigurationItem(scope_uri="", section=section)]
        )
    )

    # We should get list of configuration items here with exactly one element
    assert isinstance(config, list)
    assert len(config) == 1

    return config[0]
